# Author: Oriol Filter
# Date: 10/04/2022


## Vars

$DebugPreference = "Continue"
$ErrorActionPreference = "Stop"

### Disable Participation messages
#Set-PowerCLIConfiguration -Scope User -ParticipateInCEIP $false -Confirm:$false
#Set-PowerCLIConfiguration -Scope User -InvalidCertificateAction Ignore -Confirm:$false
#Set-PowerCLIConfiguration -Scope User -DefaultVIServerMode Multiple -Confirm:$false





## Add Vhost objects for each type of server
class _Generic_VirtualMachine
{
    [Object]$__ServerObject
    [Array]$__Snapshots=@()
    [Object]$__VMObject
    [Object]Get_ServerObject()
    {
        return $this.__ServerObject
    }
    Set_ServerObject([Object] $Server)
    {
        [Object] $this.__ServerObject=$Server
    }
    Set_VMObject([Object] $object)
    {
        $this.__VMObject=$object
    }
    [Object]Get_VMObject()
    {
        return $this.__VMObject.PowerState
    }
    [bool]__Get_Name()
    {
        $type=$this.GetType().BaseType
        Write-Warning "Get_Name not implemented in type $type"
        return $false
    }
    [String]Get_Name()
    {
        return $this.__Get_Name()
    }
    [bool]__Get_Is_Running()
    {
        $type=$this.GetType().BaseType
        Write-Warning "Get_Is_Running not implemented in type $type"
        return $false
    }
    [bool]Get_Is_Running()
    {
        return $this.__Get_Is_Running()
    }
    [Array]__Get_All_Snapshots()
    {
        $type=$this.GetType().BaseType
        Write-Warning "Get_All_Snapshots not implemented in type $type"
        return $null
    }
    [Array]Get_All_Snapshots()
    {
        return $this.__Get_Snapshots
    }
    __Load_Snapshots(){
        $type=$this.GetType().BaseType
        Write-Warning "Load_Snapshots not implemented in type $type"
    }
    Load_Snapshots(){
        $snapshots=$this.__Load_Snapshots()
        if ($snapshots){$this.__Snapshots=$snapshots}
        [String]::Format("Loaded {1} Snapshots from the VM: '{0}' at {2}",$this.Get_Name(),$this.__Snapshots.Length,$this.url) | Write-Debug
    }
    [Object]__Get_Last_Snapshot(){
        if ($this.__Snapshots.Lenght -lt 1){
            Write-Warning "No Snapshots where found, did they where loaded?"
        }
        return $this.__Snapshots[0]
    }
    [Object]Get_Last_Snapshot()
    {
        return $this.Get_Last_Snapshot()
    }
}

class VM_SCVMM:_Generic_VirtualMachine
{
    [String]__Get_Name()
    {
        return  $this.__ServerObject.Name
    }
    [bool]__Get_Is_Running()
    {
        return @{"PoweredOn"=$true;"PoweredOff"=$false}[$this.__VMObject.VirtualMachineState]
    }
    __Load_Snapshots(){
        Get-SCVMCheckpoint -VM $this.Get_Name() # -VMMServer $this.__ServerObject.url
    }

    #Get-SCVMCheckpoint
}

class VM_vSphere:_Generic_VirtualMachine
{
    [String]
    __Get_Name()
    {
        return  $this.__ServerObject.Name
    }
    [bool]
    __Get_Is_Running()
    {
        return @{ "PoweredOn" = $true; "PoweredOff" = $false }[$this.__VMObject.PowerState]
    }
    __Load_Snapshots(){
        Get-Snapshot -VM $this.Get_Name() # -VIServer $this.__ServerObject.url
    }
}

class VirtualizationServer
{
    # The fuck does all this garbage
#    [_Generic_VirtualMachine]$__vm_obj=[_Generic_VirtualMachine]
    [int]$port=""
    [string]$url=""
    [Array] $__host_list=@()
    #     [Boolean] _FindHost ($vname) {
    #         $type=$this.GetType().BaseType
    #         Write-Host "Find Host not implemented in type $type"
    #         return $false
    #     }
    #     [Boolean] FindHost ($vname) {
    #         return $this._FindHost($vname)
    #     }
#    [_Generic_VirtualMachine]__create_vm([Object]$item){
#        [_Generic_VirtualMachine]$new_vm=$this.__vm_obj::new()
#        $new_vm.__ServerObject=$this
#        $new_vm.__VMObject=$item
#        return $new_vm
#    }

    __load_hosts(){
        $type=$this.GetType().BaseType
        Write-Warning "Load hosts not implemented in type $type"
    }

    load_hosts(){
        $this.__host_list=@()
        [String]::Format("Loading VM from: {0}",$this.url) | Write-Debug
        $this.__load_hosts()
        [String]::Format("Loaded {1} VM from: {0}",$this.url,$this.__host_list.Length) | Write-Debug
    }

    [Array] get_host_list(){
        return $this.__host_list
    }

    [Boolean] __check_login(){
        $type=$this.GetType().BaseType
        Write-Warning "Check_login not implemented in type $type"
        return $false
    }

    [Boolean] check_login(){
        [String]::Format("Attempting to login into the server {0}",$this.url) | Write-Debug
#        Invoke-Command -ComputerName $this.url { $true }
        $result=$this.__check_login()
        return $result
    }

}



class HypervisorServer: VirtualizationServer
{
    [Boolean] __check_login()
    {
        $result=Invoke-Command -ComputerName $this.url { $true }
        return $result
    }
    # Lacks get VM
    #
    #     [Boolean] _FindHost ($vname) {
    #         $result=$false
    #         try
    #         {
    #             Invoke-Command -ComputerName $this.url {
    #                 Param ($name) Get-VM -Name $name
    #             } -ArgumentList $vname
    #             $result=$?
    #         }
    #         catch
    #         {
    #             Format ("{0}: {1}",$this.url,$Error[0]) | Write-Debug
    #         }
    #
    #         return $result
    #     }
}

class SystemCenterVirtualMachineManager: HypervisorServer
{
    __load_hosts()
    {
#        Get-SCVMMServer -ComputerName $this.url
        $list = Get-VMGet-SCVirtualMachine -VMMServer  $this.url
        $this.__host_list=$list
    }
}
class vSphereServer: VirtualizationServer
{
    __load_hosts()
    {
        Connect-VIServer -Server $this.url
        $list = VMware.VimAutomation.Core\Get-VM -Location $this.url
        $this.__host_list=$list
    }
    [Boolean] __check_login()
    {
        Connect-VIServer -Server $this.url
        $result=$?
#        Disconnect-VIServer -Server $this.url
        return $result
    }

}
# This remains unused??
#class VirtualizerManager
#{
#    load_hyperv_from_file(){}
#    load_scvmm_from_file(){}
#    load_vcenter_from_file(){}
#}

class VirtualizerAssistant
{
    [Array]$__connected_servers=@()
    [Hashtable]$__found_host_dict=@{}
    [Boolean] __log_in($server){
        [String]::Format("-- {0}:",$server.url) | Write-Host
        $result=$false
        try {
            $result=$server.check_login()
#            $
#            $result=Invoke-Command -ComputerName $server.url { $true }
        }
        catch [System.Management.Automation.Remoting.PSRemotingTransportException] {
            [String]::Format("Failed to log in to URL: {0}",$server.url) | Write-Warning
            [String]::Format("{0}",$Error[0]) | Write-Debug
        }
        finally {
            $message = [string]::Format(">>`tStatus {0}`n",@("NotOK","OK")[$result])
            Write-Host $message
        }
        return $result
    }
    hidden [Hashtable] get_found_host_dict() {return $this.__found_host_dict}

    append_session($type,$url)
    {
#        $type.GetType().IsValueType
        $server=$type::new()
        $server.url=$url
        $result = $this.__log_in($server)
        if ($result)
        {
            $this.__connected_servers+=$server
        }
    }
    [Array] connected_servers() {
        return $this.__connected_servers
    }

    load_hosts() {
        $this.__found_host_dict=@{}
        for ($num = 0 ; $num -le $this.__connected_servers.Length -1 ; $num++){
            $server = $this.__connected_servers[$num]
            $server.load_hosts()
            foreach ($vhost in $server.get_host_list())
            {
                $this.append_vhost($vhost,$server.url)
            }
        }
    }
    __append_vhost($vhost,$source_url){
        if ($this.__found_host_dict[$vhost.Name] -isnot [Hashtable]){
            $this.__found_host_dict[$vhost.Name]=@{}
        }
        $this.__found_host_dict[$vhost.Name][$source_url]=$vhost
    }
    append_vhost($vhost,$source_url){
        $this.__append_vhost($vhost,$source_url)
    }
    find_hosts($vname){
        $hosts_to_find=$vname
        if ($hosts_to_find -isnot [Array]){
            $hosts_to_find=@("$hosts_to_find")
        }
        foreach ($hostname  in $hosts_to_find) {
            $entry_from_list=$this.__found_host_dict[$hostname]
            [String]::Format("`t{0}:",$hostname) | Write-Host
            if ($entry_from_list.Length -gt 0)
            {
                foreach ($entry_from_dict in $entry_from_list.GetEnumerator())
                {
                    $server_location = $entry_from_dict.Name
                    [String]::Format("`t`tFound in {0}", $server_location) | Write-Host
                }
                Write-Host "`-----------`n"
            }
            else {
                [String]::Format("`----------- Couldn't be found`n") | Write-Host
            }
        }
    }
    #     find_host($vname){
    #         $hostlist=$vname
    #         if ($hostlist -is [Array]){
    #             $hostlist=@($hostlist)
    #         }
    #         foreach ($_vname in $hostlist)
    #         {
    #             for ($num = 0 ; $num -le $this.__connected_servers.Length -1 ; $num++){
    #                     $server= $this.__connected_servers[$num]
    #                     $result = $server.FindHost($_vname)
    #                     [String]::Format(">> {0} found in {1}? {2}", $_vname, $server.url, $result) | Write-Host
    #             }
    #             }
    #         }
}


#Used to store the config
class Config{
    $_SCRIPT_PATH=$PSScriptRoot
    $_HYPERV_FILE = [String]::Format("{0}/hyperv_list.txt",$this.SCRIPT_PATH)
    $_SCVMM_FILE = [String]::Format("{0}/scvmm_list.txt",$this.SCRIPT_PATH)
    $_VSPHERE_FILE = [String]::Format("{0}/vsphere_list.txt",$this.SCRIPT_PATH)
    $_HOSTS_FILE = [String]::Format("{0}/hosts.txt",$this.SCRIPT_PATH)
    Load_Hyperv($manager) {
        $hyperv_urls = Get-Hosts-From-File($this._HYPERV_FILE)
        foreach ($url in $hyperv_urls) {
            $manager.append_session([HypervisorServer], $url)
        }
    }

    Load_Scvmm([VirtualizerAssistant] $manager) {
        $hyperv_urls = Get-Hosts-From-File($this._SCVMM_FILE)
        return @{[SystemCenterVirtualMachineManager]:3}
#        foreach ($url in $hyperv_urls) {
#            $manager.append_session([SystemCenterVirtualMachineManager], $url)
#        }
    }
    Load_Vsphere($manager) {
        $vsphere_urls = Get-Hosts-From-File($this._VSPHERE_FILE)
        foreach ($url in $vsphere_urls) {
            $manager.append_session([vSphereServer], $url)
        }
    }

    [String[]]GetHostsFromFile($filepath)
    {
        ## In a future use a yaml as a .conf (with the hyperv and vmware listed)
        $file_content = Get-Content "$filepath"
        $splitted_file_content = $file_content.Split()
        $host_array=@()
        foreach ($hostname in $splitted_file_content)
        {
            if ($hostname)
            {
                $host_array+=$hostname
            }
        }
        return $host_array
    }


#    Load_All_Servers($x){
#        $this.G
#    }
#    Load_All{
#
#    }
}


class Menu{

    $exit=$false
    [VirtualizerAssistant]$Assistant
    [Config]$Config

    print_menu(){
        Write-Host "Select an option an option:"
        Write-Host @"
        `t- 0 Attempt to log in into the servers (authenticates with the current session)
        `t- 1 Load VMs
        `t- 2 Find Hosts
        `t- 3 Disconnect from all servers and reload server list content
        `t- 4 Reload VM file list content
        `t- -1 Exit
"@
    }

    select_menu(){
        $uinput=Read-Host "Please enter your option"
        switch ($uinput) {
            -1 {
                $this.Quit()
            }
            0 {

            }
            Default {
                Write-Host "Error, try again"
            }
        }
    }

    Menu(){
        while (!$this.exit){
            $this.print_menu()
            $this.select_menu()
        }
    }
    Quit(){
        $this.exit=$true
    }

}


# Advanced params?
# https://docs.microsoft.com/en-us/powershell/module/microsoft.powershell.core/about/about_functions_advanced_parameters?view=powershell-7.2

#
#function Test_VM{
#    $_____url=$_____url
#    $manager = [VirtualizerAssistant]::new()
#    $manager.append_session([SystemCenterVirtualMachineManager], $_____url)
#    $manager.__connected_servers[0].get_host_list()[0]
#}



function Main{
    load_vsphere($manager)
    $manager.load_hosts()
    $vmhosts = Get-Hosts-From-File($HOSTS_FILE)
    $manager.find_hosts($vmhosts)
#    $manager.__found_host_dict
}

Main

# if file is empty -> Debug+warn + skipping message (returns empty array)
# add vmware support
# in scvmm check if vname stored are actually the whole object or just the name
# add middleware
# do menus
# store data into custom classes, that way one with custom methods can access the information one way or another, but it's there
# add a menu()
#   get_location
#   get_status
#   list_snaps
# Replace virtual host terminology with virtual machine
# add to middleware, load hyperv_from_file($path)
# mayb load all the snapshots into the object, but unsure how it would affect the performance, afterwards can filter by name
# Check login into svcmm specifying the server (this will allow async)
# Check login vmware
# Check get VM from vcenter
# Mayb some background/Async process.
# Mayb a function/Class to load the arguments and create some conf
# Sort Snaps by date (newest to oldes)

